/*{{{}}}*/
/*{{{  Notes*/
/*                        Copyright (c) 1987 Bellcore
 *                            All Rights Reserved
 *       Permission is granted to copy or use this program, EXCEPT that it
 *       may not be sold for profit, the copyright notice must be reproduced
 *       on copies, and credit should be given to Bellcore where it is due.
 *       BELLCORE MAKES NO WARRANTY AND ACCEPTS NO LIABILITY FOR THIS PROGRAM.
 */

/* get the mouse input */
/*}}}  */

/*{{{  #includes*/
#include <unistd.h>

#include "share.h"
/*}}}  */
/*{{{  #defines*/
#define DECODE(X)	( (X) &0x80 ? (X) - 256 : (X) )
#define GET_CHAR(fd)	(index<cnt ? mouse_buff[index++] : mouse_fill(fd))
/*}}}  */

/*{{{  variables*/
#ifdef sun
/* Broken termios ?! */
static unsigned char mouse_buff[1];
#else
static unsigned char mouse_buff[90];
#endif
static int cnt=0;
static int index=0;
static int button_map[8] = { 0,1,2,3,4,5,6,7};
/*}}}  */

/*{{{  mouse_fill*/
static int mouse_fill(mouse) int mouse;
{
#ifdef MOVIE
  log_time();
#endif
#ifdef NEW_ONE
    /* this is a botch to correct for a change in read semantics for 4.1
     * that requires the mouse to in non-blocking mode
     */

    while ((cnt = read(mouse,mouse_buff,MOUSE_BUFF))<0) {
       int mask = 1<<mouse;
       select(32,&mask,0,0,NULL);
       }

   index = 1;
   return ( ((int)(*mouse_buff) & 0xff));
   }
#else
  /* That's ok, because there is a timeout on input */
  cnt=read(mouse,mouse_buff,sizeof(mouse_buff));
  index=1;
  return (cnt>0 ? *mouse_buff : 0);
}
#endif
/*}}}  */
#ifdef MOUSESYSTEMS_MOUSE
/*{{{  mouse_get*/
/* primary mouse interface
   Mouse sample begins with a byte, 0x80 through 0x87, followed by pairs of
   bytes giving motion changes.  Currently we handle 3- or 5-byte mice.
*/

int mouse_get(mouse,x_delta,y_delta) register int mouse; register int *x_delta, *y_delta;
{
  static int mouse_type = 3;
  register int delta, button;

  while(((button = GET_CHAR(mouse))&0xf0) != 0x80) if( mouse_type < 5 ) mouse_type++;
  delta = GET_CHAR(mouse); *x_delta = DECODE(delta);
  delta = GET_CHAR(mouse); *y_delta = DECODE(delta);
  if (mouse_type >= 5) 
  {
    /* 5-or-more bytes per sample */
    delta = GET_CHAR(mouse), *x_delta += DECODE(delta);
    delta = GET_CHAR(mouse), *y_delta += DECODE(delta);
  }
#ifdef MOVIE
	log_time();
#endif
  return(button_map[(~button)&0x7]);
}
/*}}}  */
#endif
#ifdef MICROSOFT_MOUSE
/*{{{  mouse_get*/
/* Andrew Haylett, 25/3/92 */
static int DELTA =16, ACCEL = 2; /* Accelerator!!!! */

int mouse_get(mouse,x_delta,y_delta) register int mouse; register int *x_delta, *y_delta;
{
  register unsigned char c1,c2,c3;
  int button;
  register int xmul,ymul;

  while(((c1 = GET_CHAR(mouse))&0xc0) != 0xc0);

  button = (c1 & 0x30)>>4;

  c2 = GET_CHAR(mouse);
  c3 = GET_CHAR(mouse);
  xmul = c1 & 0x03;
  xmul = c1 & 0x03;
  if (xmul == 0) xmul = 1;
  else if (xmul == 1) xmul = 4;
  else if (xmul == 2) xmul = -4;
  else if (xmul == 3) xmul = -1;
  ymul = (c1 >> 2) & 0x03;
  if (ymul == 0) ymul = 1;
  else if (ymul == 1) ymul = 8;
  else if (ymul == 2) ymul = -8;
  else if (ymul == 3) ymul = -1;
  *x_delta = (xmul < 0) ? ((~c2 & 0x3f) + 1) : c2 & 0x3f;
  *y_delta = (ymul < 0) ? ((~c3 & 0x3f) + 1) : c3 & 0x3f;
  if (*x_delta > DELTA) xmul *= ACCEL;
  if (*y_delta > DELTA) ymul *= ACCEL;
  *x_delta *= xmul;
  *y_delta *= -ymul;

#if 1
   if (button == 2)
     button = 4;
   else if (button == 3)
     button = 2;
#else
   button =((button>>1)&0x01)|((button<<2)&0x04);/* Map buttons accordingly */
#endif
#ifdef MOVIE
	log_time();
#endif
   return(button_map[(button)&0x7]);
}
/*}}}  */
#endif
/*{{{  map_mouse buttons (for non-left handers)*/
int *map_mouse(button,map) int button, map;
{
  if (button>0 && button<8) button_map[button]=map;
  return(button_map);
}
/*}}}  */
/*{{{  mouse_count -- are there still characters in the mouse buffer?*/
int mouse_count()
{
  return(cnt>index);
}
/*}}}  */
